/*  SciGraphica - Scientific graphics and data manipulation
 *  Copyright (C) 2000 Adrian E. Feiguin <feiguin@ifir.edu.ar>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gtkextra/gtkextra.h>

#include "main.h"

enum{
  IMPORT_WORKSHEET,
  EXPORT_WORKSHEET,
  IMPORT_PLOT,
  EXPORT_PLOT,
  IMPORT_PROJECT,
  EXPORT_PROJECT,
};

static gchar *titles[] = {
                       "Import Worksheet",
                       "Export Worksheet",
                       "Import Plot",
                       "Export Plot",
                       "Read Project",
                       "Save Project"};

static gchar *project_export_formats[] = {
                       "SciGraphica XML (*.sg)",
                       NULL,};
static gchar *project_export_filters[] = {
                       "*.sg",
                       NULL,};
static gchar *project_import_formats[] = {
                       "SciGraphica XML (*.sg)",
                       NULL,};
static gchar *project_import_filters[] = {
                       "*.sg",
                       NULL,};
static gchar *worksheet_export_formats[] = {
                       "SciGraphica Worksheet (*.sgw)",
                       "ASCII",
                       "HTML",
                       "TeX",
                       NULL,};
static gchar *worksheet_export_filters[] = {
                       "*.sgw",
                       "*.dat",
                       "*.htm*",
                       "*.tex",
                       NULL,};
static gchar *worksheet_import_formats[] = {
                       "SciGraphica Worksheet XML (*.sgw)",
                       "ASCII",
                       NULL,};
static gchar *worksheet_import_filters[] = {
                       "*.sgw",
                       "*.dat",
                       NULL,};
static gchar *plot_export_formats[] = {
                       "PostScript (*.ps)",
                       "Encapsulated PostScript (*.eps)",
                       "Scigraphica Plot (*.sgp)",
                       NULL,};
static gchar *plot_export_filters[] = {
                       "*.ps",
                       "*.eps",
                       "*.sgp",
                       NULL,};
static gchar *plot_import_formats[] = {
                       "Scigraphica Plot XML (*.sgp)",
                       NULL,};
static gchar *plot_import_filters[] = {
                       "*.sgp",
                       NULL,};

static GtkWidget *format_combo;
static gint action;

static void select_file			();
static gint cancel_action 		(GtkWidget *widget, gpointer data);
static gint ok_clicked 			(GtkWidget *widget, gpointer data);
static void new_format			(GtkWidget *widget, gpointer data);
static void set_path			(gchar **path, gchar *new_path);
static gboolean	check_file		(gchar *path);

void
sg_project_import()
{
  action = IMPORT_PROJECT;
  select_file();
}

void
sg_project_export()
{
  action = EXPORT_PROJECT;
  select_file();
}


void
sg_worksheet_import()
{
  action = IMPORT_WORKSHEET;
  select_file();
}

void
sg_worksheet_export()
{
  action = EXPORT_WORKSHEET;
  select_file();
}

void
sg_plot_export()
{
  action = EXPORT_PLOT;
  select_file();
}

void
sg_plot_import()
{
  action = IMPORT_PLOT;
  select_file();
}


static void
select_file()
{
  GtkWidget *filesel;
  GtkWidget *label;
  gchar **labels;
  gchar *path = NULL;

  filesel = gtk_icon_file_selection_new (titles[action]);

  gtk_table_resize(GTK_TABLE(GTK_ICON_FILESEL(filesel)->action_area), 3, 4);

  label = gtk_label_new("File Format:        ");
  gtk_misc_set_alignment(GTK_MISC(label), 1., 0.5);
  gtk_table_attach_defaults(GTK_TABLE(GTK_ICON_FILESEL(filesel)->action_area),
                            label,
                            0, 1, 2, 3);
  gtk_widget_show(label);

  format_combo = gtk_combo_new();
  gtk_table_attach_defaults(GTK_TABLE(GTK_ICON_FILESEL(filesel)->action_area),
                            format_combo,
                            1, 3, 2, 3);
  gtk_widget_show(format_combo);
  gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(format_combo)->entry), FALSE);

  switch(action){
    case IMPORT_WORKSHEET:
      path = last_worksheet_path;
      labels = worksheet_import_formats;
      break;
    case EXPORT_WORKSHEET:
      path = last_worksheet_path;
      labels = worksheet_export_formats;
      break;
    case EXPORT_PLOT:
      path = last_plot_path;
      labels = plot_export_formats;
      break;
    case IMPORT_PLOT:
      path = last_plot_path;
      labels = plot_import_formats;
      break;
    case EXPORT_PROJECT:
      path = last_project_path;
      labels = project_export_formats;
      break;
    case IMPORT_PROJECT:
      path = last_project_path;
      labels = project_import_formats;
      break;
    default:
      labels = NULL;
  }

  gtk_dir_tree_open_dir(GTK_DIR_TREE(GTK_ICON_FILESEL(filesel)->dir_tree), path);

  gtk_window_position (GTK_WINDOW (filesel), GTK_WIN_POS_CENTER);

  /* Connect the ok_button function */
  gtk_signal_connect (GTK_OBJECT (GTK_ICON_FILESEL(filesel)->ok_button),
                      "clicked",
                      (GtkSignalFunc) ok_clicked, filesel);

  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect (GTK_OBJECT (GTK_ICON_FILESEL(filesel)->cancel_button),
                      "clicked",
                      (GtkSignalFunc) cancel_action, filesel);

  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect (GTK_OBJECT (GTK_COMBO(format_combo)->entry),
                      "changed",
                      (GtkSignalFunc) new_format, filesel);

  sg_combo_set_items(GTK_COMBO(format_combo), labels);

  gtk_widget_show_all(filesel);


}

static void
new_format(GtkWidget *widget, gpointer data)
{
  GtkWidget *child;
  gint file_format;
  gchar *filter = NULL;

  child = GTK_LIST(GTK_COMBO(format_combo)->list)->selection->data;
  file_format = gtk_list_child_position(GTK_LIST(GTK_COMBO(format_combo)->list), child);

  switch(action){
    case IMPORT_WORKSHEET:
      filter = worksheet_import_filters[file_format];
      break;
    case EXPORT_WORKSHEET:
      filter = worksheet_export_filters[file_format];
      break;
    case EXPORT_PLOT:
      filter = plot_export_filters[file_format];
      break;
    case IMPORT_PLOT:
      filter = plot_import_filters[file_format];
      break;
    case EXPORT_PROJECT:
      filter = project_export_filters[file_format];
      break;
    case IMPORT_PROJECT:
      filter = project_import_filters[file_format];
      break;
  }


  gtk_entry_set_text(GTK_ENTRY(GTK_ICON_FILESEL(data)->filter_entry), filter); 
  gtk_file_list_set_filter(GTK_FILE_LIST(GTK_ICON_FILESEL(data)->file_list), filter);

}


static gint 
cancel_action (GtkWidget *widget, gpointer data) {
  gtk_widget_destroy(GTK_WIDGET(data));
  return FALSE; 
}

static gint
ok_clicked (GtkWidget *widget, gpointer data)
{
  GtkSheetRange range;
  GtkIconFileSel *filesel;
  GtkWidget *child = NULL;
  gchar *file_path, *file_name, *full_path;
  gint file_format;
  gboolean veto = TRUE;
  gboolean file_exists = FALSE;
  gchar message[255];

  filesel = GTK_ICON_FILESEL(data);
  file_name = gtk_file_list_get_filename(GTK_FILE_LIST(filesel->file_list));
  file_path = gtk_file_list_get_path(GTK_FILE_LIST(filesel->file_list));

  if(!file_name){
     file_name = gtk_entry_get_text(GTK_ENTRY(filesel->file_entry));
  }

  full_path = g_strdup(file_path);
  full_path = g_strconcat(full_path, file_name, NULL);

  child = GTK_LIST(GTK_COMBO(format_combo)->list)->selection->data;
  file_format = gtk_list_child_position(GTK_LIST(GTK_COMBO(format_combo)->list), child);

  file_exists = check_file(full_path);
  if(file_exists && 
     (action == EXPORT_WORKSHEET || action == EXPORT_PLOT ||
      action == EXPORT_PROJECT)){
        sprintf(message, "File Exists. Overwrite %s?", file_name);
        veto = sg_accept_dialog(message);
        if(!veto){
            return TRUE;
        }
  }
  if(!file_exists && 
     (action == IMPORT_WORKSHEET || action == IMPORT_PLOT ||
      action == IMPORT_PROJECT)){
        sg_message_dialog("File not found", 0);
        return TRUE;
  }


  switch(action){
    case IMPORT_WORKSHEET:
      set_path(&last_worksheet_path, file_path);
      set_path(&last_worksheet_filename, file_name);
      switch(file_format){
        case 0:
          sg_project_file_import_xml(full_path, active_worksheet, NULL);
          break;
        case 1:
          sg_worksheet_file_import_ascii(active_worksheet, full_path);
          break;
      }

      if(rename_worksheets) {
        sg_worksheet_rename(active_worksheet, file_name); 
      }

      gtk_icon_list_set_active_icon(GTK_ICON_LIST(iconlist1), 
                                    active_worksheet->icon);
      gtk_icon_list_set_active_icon(GTK_ICON_LIST(iconlist1), NULL);

      break;
    case EXPORT_WORKSHEET:
      set_path(&last_worksheet_path, file_path);
      set_path(&last_worksheet_filename, file_name);
      switch(file_format){
        case 0:
           range.row0 = range.col0 = 0;
           range.coli = GTK_SHEET(active_worksheet->sheet)->maxcol;
           range.rowi = GTK_SHEET(active_worksheet->sheet)->maxrow;
/*
           full_path = g_strconcat(full_path, ".sgw", NULL);
*/
           sg_worksheet_file_export_xml(active_worksheet, full_path, &range,
                                        NULL);
           break;
        case 1:
           sg_worksheet_file_export_ascii(active_worksheet, full_path, NULL);
           break;
        case 2:
           sg_worksheet_file_export_html(active_worksheet, full_path, NULL);
           break;
        case 3:
           sg_worksheet_file_export_tex(active_worksheet, full_path, NULL);
           break;
      }
      break;
    case EXPORT_PLOT:
      set_path(&last_plot_path, file_path);
      set_path(&last_plot_filename, file_name);
      switch(file_format){
        case 2:
           sg_plot_file_export_xml(active_plot, full_path, NULL);
           break; 
        case 1:
           gtk_plot_canvas_export_ps(GTK_PLOT_CANVAS(active_plot->real_canvas),
                                     full_path, 
                                     active_plot->orientation,
                                     TRUE,
                                     active_plot->page_size);
           break;
        case 0:
        default:
           gtk_plot_canvas_export_ps(GTK_PLOT_CANVAS(active_plot->real_canvas),
                                     full_path, 
                                     active_plot->orientation,
                                     FALSE,
                                     active_plot->page_size);
           break;
      }
      break;
    case IMPORT_PLOT:
      set_path(&last_plot_path, file_path);
      set_path(&last_plot_filename, file_name);
      switch(file_format){
        case 0:
          sg_project_file_import_xml(full_path, NULL, active_plot);
          break;
      }

      if(rename_plots) {
        sg_plot_rename(active_plot, file_name); 
      }

      gtk_icon_list_set_active_icon(GTK_ICON_LIST(iconlist2), 
                                    active_plot->icon);
      gtk_icon_list_set_active_icon(GTK_ICON_LIST(iconlist2), NULL);

      break;
    case EXPORT_PROJECT:
      set_path(&last_project_path, file_path);
      set_path(&last_project_filename, file_name);
      sg_project_file_export_xml(full_path);
      break;
    case IMPORT_PROJECT:
      set_path(&last_project_path, file_path);
      set_path(&last_project_filename, file_name);
      sg_project_file_import_xml(full_path, NULL, NULL);
      break;
  }

  gtk_widget_destroy(GTK_WIDGET(data));
  g_free(full_path);
  return TRUE;
}

static void
set_path(gchar **path, gchar *new_path)
{
  if(*path)
    g_free(*path);

  *path = g_strdup(new_path);
}

static gboolean 
check_file(gchar *path)
{
  FILE *fp;
  gboolean file_exists = FALSE;

  if((fp = fopen(path, "r"))){
     fclose(fp);
     file_exists = TRUE;
  } 

  return file_exists;
}
