/*  SciGraphica - Scientific graphics and data manipulation
 *  Copyright (C) 2000 Adrian E. Feiguin <feiguin@ifir.edu.ar>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gtkextra/gtkextra.h>
#include <parser.h>
#include <stdlib.h>

#include "main.h"

/* TODO
 * dataset function
 */

enum {
	PARSE_WORKSHEET,
	PARSE_PLOT,
	PARSE_PROJECT,
};

enum {
	PARSER_IN_WORKSHEET,
	PARSER_IN_PLOT,
	PARSER_IN_PAGE,
	PARSER_IN_LAYER,
	PARSER_IN_AXIS,
	PARSER_IN_TITLE,
	PARSER_IN_VMAJOR,
	PARSER_IN_VMINOR,
	PARSER_IN_HMAJOR,
	PARSER_IN_HMINOR,
	PARSER_IN_X0,
	PARSER_IN_Y0,
	PARSER_IN_LEGENDS,
	PARSER_IN_DATA,
	PARSER_IN_XLINE,
	PARSER_IN_YLINE,
	PARSER_IN_POINTS,
	PARSER_UNKNOWN,
};

typedef struct _sgParseState sgParseState;
struct _sgParseState {
  gchar *name;
  GString *content;

  gint object;

  gint state;

  gint num_layers;

  gboolean in_text;

  SGworksheet *worksheet;
  gint row, col;

  SGplot *plot;
  SGlayer *layer;
  SGdataset *dataset;
  GtkPlotText text;
  gint axis;
};

static void sgWarning				(sgParseState *state, 
						 const char *msg, ...);
static void sgError				(sgParseState *state, 
						 const char *msg, ...);
static void sgFatalError			(sgParseState *state, 
						 const char *msg, ...);
static void sgStartDocument			(sgParseState *state);
static void sgEndDocument			(sgParseState *state);
static void sgCharacters			(sgParseState *state, 
						 const CHAR *chars,
                                                 int len); 
static void sgStartElement			(sgParseState *state, 
						 const CHAR *name, 
						 const CHAR **attrs);
static void sgEndElement			(sgParseState *state, 
						 const CHAR *name); 

static const CHAR *get_real_name		(const CHAR *name);


static xmlSAXHandler sgSAXParser = {
  0, /* internal Subset */
  0, /* isStandalone */
  0, /* hasInternalSubset */
  0, /* hasExternalSubset */
  0, /* resolveEntity */
  0, /* getEntity */
  0, /* entityDecl */
  0, /* notationDecl */
  0, /* attributeDecl */
  0, /* elementDecl */
  0, /* unparseEntityDecl */
  0, /* setDocumentLocator */
  (startDocumentSAXFunc)sgStartDocument, /* startDocument */
  (endDocumentSAXFunc)sgEndDocument, /* endDocument */
  (startElementSAXFunc)sgStartElement, /* startElement */
  (endElementSAXFunc)sgEndElement, /* endElement */
  0, /* reference */
  (charactersSAXFunc)sgCharacters, /*characters */
  0, /* ignorableWhitespace */
  0, /* processingInstruction */
  (commentSAXFunc)0, /* comment */
  (warningSAXFunc)sgWarning, /* warning */
  (errorSAXFunc)sgError, /* error */
  (fatalErrorSAXFunc)sgFatalError, /* fatalError */
};

static void  
sgWarning(sgParseState *state, const char *msg, ...)
{
  va_list args;

  va_start(args, msg);
  g_logv("XML", G_LOG_LEVEL_WARNING, msg, args);
  va_end(args);
}

static void  
sgError(sgParseState *state, const char *msg, ...)
{
  va_list args;

  va_start(args, msg);
  g_logv("XML", G_LOG_LEVEL_CRITICAL, msg, args);
  va_end(args);
}

static void  
sgFatalError(sgParseState *state, const char *msg, ...)
{
  va_list args;

  va_start(args, msg);
  g_logv("XML", G_LOG_LEVEL_ERROR, msg, args);
  va_end(args);
}

static void
sgStartDocument(sgParseState *state)
{
  state->name = NULL;
  state->content = g_string_sized_new(128);

  state->state = PARSER_UNKNOWN;

}

static void
sgEndDocument(sgParseState *state)
{
  if(state->text.font)
   g_free(state->text.font);

  if(state->name)
   g_free(state->name);

  if(state->content)
   g_string_free(state->content, TRUE);
}

static void
sgCharacters(sgParseState *state, const CHAR *chars, int len)
{
  gint i;

  for(i = 0; i < len; i++){
    g_string_append_c(state->content, chars[i]);    
  }
}

static void
sgStartElement(sgParseState *state, const CHAR *real_name, const CHAR **attrs)
{
  SGplot *plot = state->plot;
  GtkPlot *real_plot = NULL;
  GtkPlotCanvas *canvas = NULL;
  GtkSheet *sheet = NULL;
  const CHAR **aux = attrs;
  const CHAR *child = NULL, *value = NULL;
  const CHAR *name;

  name = get_real_name(real_name);

  if(state->plot) {
    canvas = GTK_PLOT_CANVAS(state->plot->real_canvas);
  }

  if(state->worksheet) {
    sheet = GTK_SHEET(state->worksheet->sheet);
  }

  state->name = g_strdup(real_name);
  state->content = g_string_truncate(state->content, 0);

  if(state->layer) real_plot = GTK_PLOT(state->layer->real_plot);

  if(strcmp(name, "Worksheet") == 0){
     state->state = PARSER_IN_WORKSHEET;

     state->row = 0;
     state->col = 0;

     state->plot = NULL;
     state->layer = NULL;

     if(state->object == PARSE_PROJECT){
         state->worksheet = sg_project_new_worksheet();
         sheet = GTK_SHEET(state->worksheet->sheet);
     }
  }

  if(strcmp(name, "Plot") == 0){
     state->state = PARSER_IN_PLOT;
     state->in_text = FALSE;

     state->worksheet = NULL;

     if(state->object == PARSE_PROJECT){
         state->plot = sg_project_new_plot();
         canvas = GTK_PLOT_CANVAS(state->plot->real_canvas);
     }
  }

  if(strcmp(name, "Text") == 0){
     state->in_text = TRUE;
  }

  if(state->state != PARSER_IN_LAYER && strcmp(name, "Geometry") == 0){
     gint width = 400;
     gint height = 350;

     while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Width") == 0) width = atoi(value); 
        if(strcmp(child, "Height") == 0) height = atoi(value); 
     }

     if(state->state == PARSER_IN_PLOT){
       state->plot->width = width; 
       state->plot->height = height; 
       if(state->plot->is_mapped)
             gtk_widget_set_usize(plot->window, width, height);
     }

     if(state->state == PARSER_IN_WORKSHEET){
       state->worksheet->width = width;
       state->worksheet->height = height;
       if(state->worksheet->is_mapped)
            gtk_widget_set_usize(state->worksheet->window, width, height);
     }

  }

  if(state->state == PARSER_IN_WORKSHEET && strcmp(name, "Column") == 0){
     gint ncol = 0;
     gint width = 80;
     const gchar *title = NULL;
     gint format = 0;
     gint precision = 3;

     while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Width") == 0) width = atoi(value);
        if(strcmp(child, "No") == 0) ncol = atoi(value);
        if(strcmp(child, "Title") == 0) title = value;
        if(strcmp(child, "Format") == 0) format = atoi(value);
        if(strcmp(child, "Presicion") == 0) precision = atoi(value);
     }
     gtk_sheet_set_column_width(sheet, ncol, width);
     gtk_sheet_set_column_title(sheet, ncol, title);
     gtk_sheet_column_button_add_label(sheet, ncol, title);
     sg_worksheet_column_set_format(state->worksheet, ncol, format, precision);
     state->col = ncol;
  }

  if(state->state == PARSER_IN_WORKSHEET && strcmp(name, "Cell") == 0){
     while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Row") == 0) state->row = atoi(value);
        if(strcmp(child, "Col") == 0) state->col = atoi(value);
     }
  }


  if(strcmp(name, "Page") == 0){
     gint width = GTK_PLOT_LETTER_W;
     gint height = GTK_PLOT_LETTER_H;
     gint orientation = GTK_PLOT_PORTRAIT;
     gint page_size = GTK_PLOT_LETTER;
     gdouble scale = .65;

     state->state = PARSER_IN_PAGE;
     state->in_text = FALSE;

     while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Width") == 0) width = atoi(value); 
        if(strcmp(child, "Height") == 0) height = atoi(value); 
        if(strcmp(child, "Orientation") == 0) orientation = atoi(value); 
        if(strcmp(child, "Size") == 0) page_size = atoi(value); 
        if(strcmp(child, "Scale") == 0) scale = atof(value); 
     }
     sg_plot_rescale(plot, scale);
     sg_plot_set_size(plot, page_size, width, height, orientation);
  }

  if(strcmp(name, "Layer") == 0){
     state->state = PARSER_IN_LAYER;
     state->in_text = FALSE;

     state->num_layers++;
  }

  if(state->state == PARSER_IN_LAYER && strcmp(name, "Geometry") == 0){
     GList *list = NULL;
     gdouble x = .175;
     gdouble y = .15;
     gdouble width = .65;
     gdouble height = .45;
     gint xscale = 0;
     gint yscale = 0;
     gint num_plots = 0;

     while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "X") == 0) x = atof(value); 
        if(strcmp(child, "Y") == 0) y = atof(value); 
        if(strcmp(child, "Width") == 0) width = atof(value); 
        if(strcmp(child, "Height") == 0) height = atof(value); 
        if(strcmp(child, "Xscale") == 0) xscale = atoi(value); 
        if(strcmp(child, "Yscale") == 0) yscale = atoi(value); 
     }

/*   This is to fix the bug in gtkplotcanvas in gtkextra-0.99.6 */
     list = canvas->plots;
     while(list){
         num_plots++;
         list = list->next;
     }
 
     if(state->num_layers <= num_plots){
       state->layer = (SGlayer*) g_list_nth_data(plot->layers, 
                                                 state->num_layers - 1);
       real_plot = GTK_PLOT(state->layer->real_plot);
       gtk_plot_move_resize(real_plot, 
                            x, y, width, height);
     }
     else
     {
       state->layer = sg_layer_new(width, height);
       sg_plot_add_layer(plot, state->layer, x, y);
       real_plot = GTK_PLOT(state->layer->real_plot);
     }

     gtk_plot_set_xscale(real_plot, xscale);
     gtk_plot_set_yscale(real_plot, yscale);
  }

  if(state->state == PARSER_IN_LAYER){
    if(strcmp(name, "Range") == 0){
      gdouble xmin = 0.0;
      gdouble xmax = 1.0;
      gdouble ymin = 0.0;
      gdouble ymax = 1.0;

      while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Xmin") == 0) xmin = atof(value); 
        if(strcmp(child, "Xmax") == 0) xmax = atof(value); 
        if(strcmp(child, "Ymin") == 0) ymin = atof(value); 
        if(strcmp(child, "Ymax") == 0) ymax = atof(value); 
      }
 
      gtk_plot_set_range(real_plot, 
                         xmin, xmax, ymin, ymax);

    }


    if(strcmp(name, "Vmajor") == 0){ 

      state->state = PARSER_IN_VMAJOR;
      state->in_text = FALSE;

      while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Visible") == 0) real_plot->show_vmajor = atoi(value); 
      }

    }
    if(strcmp(name, "Vminor") == 0){ 

      state->state = PARSER_IN_VMINOR;
      state->in_text = FALSE;

      while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Visible") == 0) real_plot->show_vminor = atoi(value); 
      }

    }
    if(strcmp(name, "Hmajor") == 0){ 

      state->state = PARSER_IN_HMAJOR;
      state->in_text = FALSE;

      while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Visible") == 0) real_plot->show_hmajor = atoi(value); 
      }

    }
    if(strcmp(name, "Hminor") == 0){ 

      state->state = PARSER_IN_HMINOR;
      state->in_text = FALSE;

      while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Visible") == 0) real_plot->show_hminor = atoi(value); 
      }

    }
    if(strcmp(name, "X0") == 0){
      gint visible = FALSE;

      state->state = PARSER_IN_X0;
      state->in_text = FALSE;

      while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Visible") == 0) visible = atoi(value);
      }
      gtk_plot_x0_set_visible(real_plot, visible);
    }
    if(strcmp(name, "Y0") == 0){
      gint visible = FALSE;

      state->state = PARSER_IN_Y0;
      state->in_text = FALSE;

      while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Visible") == 0) visible = atoi(value);
      }
      gtk_plot_y0_set_visible(real_plot, visible);
    }

    if(strcmp(name, "Xticks") == 0 || strcmp(name, "Yticks") == 0){
      gdouble step = .1;
      gint nminor = 1;
      gboolean limits = FALSE;
      gdouble begin = 0.;
      gdouble end = 0.;
      gint orientation = 0;

      if(strcmp(name, "Xticks") == 0)
        orientation = GTK_ORIENTATION_HORIZONTAL;
      else
        orientation = GTK_ORIENTATION_VERTICAL;

      while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Step") == 0) step = atof(value); 
        if(strcmp(child, "Begin") == 0) begin = atof(value); 
        if(strcmp(child, "End") == 0) end = atof(value); 
        if(strcmp(child, "Nminor") == 0) nminor = atoi(value); 
        if(strcmp(child, "Limits") == 0) limits = atoi(value); 
      }

      gtk_plot_axis_set_ticks(real_plot, orientation, step, nminor);
      if(limits)
         gtk_plot_axis_set_ticks_limits(real_plot, orientation, begin, end);
      else
         gtk_plot_axis_unset_ticks_limits(real_plot, orientation); 
    }
  }

  if(state->state == PARSER_IN_LAYER && strcmp(name, "Axis") == 0){
     gdouble align = 0.;
     gboolean visible = TRUE;
     gint scale = 0;
     GtkPlotAxis *the_axis;

     state->state = PARSER_IN_AXIS;
     state->in_text = FALSE;

     while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Position") == 0){
           if(strcmp(value, "Top") == 0) state->axis = GTK_PLOT_AXIS_TOP; 
           if(strcmp(value, "Bottom") == 0) state->axis = GTK_PLOT_AXIS_BOTTOM; 
           if(strcmp(value, "Left") == 0) state->axis = GTK_PLOT_AXIS_LEFT; 
           if(strcmp(value, "Right") == 0) state->axis = GTK_PLOT_AXIS_RIGHT; 
        }
        if(strcmp(child, "Visible") == 0) visible = atoi(value);
        if(strcmp(child, "Scale") == 0) scale = atoi(value);
        if(strcmp(child, "Align") == 0) align = atof(value);
     }
   
     the_axis = gtk_plot_get_axis(real_plot, state->axis); 
     the_axis->scale_type = scale;
     gtk_plot_axis_set_visible(real_plot, state->axis, visible); 
     switch(state->axis){
       case GTK_PLOT_AXIS_TOP:
           real_plot->top_align = align;
           break;
       case GTK_PLOT_AXIS_BOTTOM:
           real_plot->bottom_align = align;
           break;
       case GTK_PLOT_AXIS_LEFT:
           real_plot->left_align = align;
           break;
       case GTK_PLOT_AXIS_RIGHT:
           real_plot->right_align = align;
           break;
     }
  }

  if(state->state == PARSER_IN_AXIS){
    if(strcmp(name, "Ticks") == 0){
      gint major_mask = 0;
      gint minor_mask = 0;
      gint length = 0, width = 0;

      state->in_text = FALSE;

      while(*aux){
         child = *aux++;
         value = *aux++;
         if(strcmp(child, "Major") == 0) major_mask = atoi(value); 
         if(strcmp(child, "Minor") == 0) minor_mask = atoi(value); 
         if(strcmp(child, "Length") == 0) length = atoi(value); 
         if(strcmp(child, "Width") == 0) width = atoi(value); 
      }
 
      gtk_plot_axis_show_ticks(real_plot, state->axis, major_mask, minor_mask);
      gtk_plot_axis_set_ticks_length(real_plot, state->axis, length);
      gtk_plot_axis_set_ticks_width(real_plot, state->axis, width);
 
    }

    if(strcmp(name, "Title") == 0){
      gboolean visible = FALSE;

      state->state = PARSER_IN_TITLE;
      state->in_text = FALSE;

      while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Visible") == 0) visible = atoi(value);
      }
      if(visible)
        gtk_plot_axis_show_title(real_plot, state->axis);
      else
        gtk_plot_axis_hide_title(real_plot, state->axis);
    }
   
    if(strcmp(name, "Labels") == 0){
      gint mask = 0;
      gint precision = 2;
      gint offset = 10;
      gint style = 0;
      GtkPlotAxis *the_axis;

      while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Mask") == 0) mask = atoi(value);
        if(strcmp(child, "Precision") == 0) precision = atoi(value);
        if(strcmp(child, "Offset") == 0) offset = atoi(value);
        if(strcmp(child, "Style") == 0) style = atoi(value);
      }
      
      the_axis = gtk_plot_get_axis(real_plot, state->axis); 
      the_axis->labels_offset = offset;
      gtk_plot_axis_show_labels(real_plot, state->axis, mask);
      gtk_plot_axis_labels_set_numbers(real_plot, state->axis, 
                                       style, precision);
    }
  }

  if(state->state == PARSER_IN_LAYER && strcmp(name, "Legends") == 0){
     gdouble x = 0.6;
     gdouble y = 0.6;
     gint width = 10.;
     gint height = 10.;
     gboolean visible = TRUE;

     state->state = PARSER_IN_LEGENDS;
     state->in_text = FALSE;

     while(*aux){
        child = *aux++;
        value = *aux++;
        if(strcmp(child, "Visible") == 0) visible = atoi(value);
        if(strcmp(child, "X") == 0) x = atof(value);
        if(strcmp(child, "Y") == 0) y = atof(value);
        if(strcmp(child, "Width") == 0) width = atoi(value);
        if(strcmp(child, "Height") == 0) height = atoi(value);
     }

     if(visible)
       gtk_plot_show_legends(real_plot);  
     else
       gtk_plot_hide_legends(real_plot);  
     gtk_plot_legends_move(real_plot, x, y); 
  } 
  
  if(strcmp(name, "Border") == 0){
    gint style = 0;
    gint width = 0;
    gint shadow = 0;

    while(*aux){
       child = *aux++;
       value = *aux++;
       if(strcmp(child, "Style") == 0) style = atoi(value); 
       if(strcmp(child, "Line") == 0) width = atoi(value); 
       if(strcmp(child, "Shadow") == 0) shadow = atoi(value); 
    }

    if(state->in_text){
        state->text.border = style; 
        state->text.border_width = width;
        state->text.shadow_width = shadow; 
    } else {
        gtk_plot_set_legends_border(real_plot, style, shadow);
        real_plot->legends_border_width = width;
    }
  }

  if(strcmp(name, "Dataset") == 0){
    gboolean visible = TRUE;
    gboolean show_labels = TRUE;
    gint connector = 0;
    const gchar *name = NULL;

    state->state = PARSER_IN_DATA;
    state->in_text = FALSE;

    while(*aux){
       child = *aux++;
       value = *aux++;
       if(strcmp(child, "Visible") == 0) visible = atoi(value); 
       if(strcmp(child, "Labels") == 0) show_labels = atoi(value); 
       if(strcmp(child, "Connector") == 0) connector = atoi(value); 
       if(strcmp(child, "Name") == 0) name = value; 
    }

    state->dataset = sg_dataset_new();

    if(!visible)
       gtk_plot_data_hide(state->dataset->real_data);

    gtk_plot_data_show_labels(state->dataset->real_data, show_labels);
    gtk_plot_data_set_connector(state->dataset->real_data, connector);

    if(name)
       sg_dataset_set_name(state->dataset, (gchar *)name);
  }  
  
  if(state->state == PARSER_IN_DATA){
    if(strcmp(name, "Legend") == 0){
      gboolean visible = TRUE;

      while(*aux){
         child = *aux++;
         value = *aux++;
         if(strcmp(child, "Visible") == 0) visible = atoi(value); 
      }

      if(!visible)
        gtk_plot_data_hide_legend(state->dataset->real_data);
    }

    if(strcmp(name, "Datasheet") == 0){
      const gchar *name = NULL;
      gint x = -1;
      gint y = -1;
      gint dx = -1;
      gint dy = -1;
      gint labels = -1;
      SGworksheet *datasheet;

      while(*aux){
         child = *aux++;
         value = *aux++;
         if(strcmp(child, "Name") == 0) name = value; 
         if(strcmp(child, "X") == 0) x = atoi(value); 
         if(strcmp(child, "Y") == 0) y = atoi(value); 
         if(strcmp(child, "DX") == 0) dx = atoi(value); 
         if(strcmp(child, "DX") == 0) dy = atoi(value); 
         if(strcmp(child, "Labels") == 0) labels = atoi(value); 
      }

      datasheet = sg_worksheet_get_by_name((gchar *)name);
      if(datasheet){
        state->dataset->worksheet = datasheet;
        state->dataset->x = x;
        state->dataset->y = y;
        state->dataset->dx = dx;
        state->dataset->dy = dy;
        state->dataset->l = labels;
      }
    }

    if(strcmp(name, "Xline") == 0){

      state->state = PARSER_IN_XLINE;
      state->in_text = FALSE;

    }

    if(strcmp(name, "Yline") == 0){

      state->state = PARSER_IN_YLINE;
      state->in_text = FALSE;

    }

  }

  if(state->state == PARSER_IN_DATA && strcmp(name, "Points") == 0){
    state->state = PARSER_IN_POINTS;
    state->in_text = FALSE;
  }

  if(state->state == PARSER_IN_POINTS && strcmp(name, "Point") == 0){
    gdouble x = 0.;
    gdouble y = 0.;
    gdouble dx = 0.;
    gdouble dy = 0.;
 
    while(*aux){
       child = *aux++;
       value = *aux++;

       if(strcmp(child, "X") == 0) x = atof(value); 
       if(strcmp(child, "Y") == 0) y = atof(value); 
       if(strcmp(child, "DX") == 0) dx = atof(value); 
       if(strcmp(child, "DY") == 0) dy = atof(value); 
    }
    sg_dataset_add_point(state->dataset, &x, &y, &dx, &dy, NULL);
  }

  if(state->state == PARSER_IN_DATA && strcmp(name, "Xerrbars") == 0){
    gboolean visible = FALSE;
    gint width = 0;
    gint caps = 0;

    while(*aux){
       child = *aux++;
       value = *aux++;

       if(strcmp(child, "Visible") == 0) visible = atoi(value); 
       if(strcmp(child, "Width") == 0) width = atoi(value); 
       if(strcmp(child, "Caps") == 0) caps = atoi(value); 
    }

    if(visible)
       gtk_plot_data_show_xerrbars(state->dataset->real_data);
    state->dataset->real_data->xerrbar_width = width;
    state->dataset->real_data->xerrbar_caps = caps;
  }

  if(state->state == PARSER_IN_DATA && strcmp(name, "Yerrbars") == 0){
    gboolean visible = FALSE;
    gint width = 0;
    gint caps = 0;

    while(*aux){
       child = *aux++;
       value = *aux++;

       if(strcmp(child, "Visible") == 0) visible = atoi(value); 
       if(strcmp(child, "Width") == 0) width = atoi(value); 
       if(strcmp(child, "Caps") == 0) caps = atoi(value); 
    }

    if(visible)
       gtk_plot_data_show_yerrbars(state->dataset->real_data);
    state->dataset->real_data->yerrbar_width = width;
    state->dataset->real_data->yerrbar_caps = caps;
  }

  if(state->state == PARSER_IN_DATA && strcmp(name, "Datalabels") == 0){
    gboolean visible = FALSE;
    gint offset = 6;

    state->in_text = FALSE;

    while(*aux){
       child = *aux++;
       value = *aux++;

       if(strcmp(child, "Visible") == 0) visible = atoi(value); 
       if(strcmp(child, "Offset") == 0) offset = atoi(value); 
    }

    gtk_plot_data_show_labels(state->dataset->real_data, visible);
    state->dataset->real_data->labels_offset = offset;
  }


  if(state->state == PARSER_IN_DATA && strcmp(name, "Label") == 0){
    const gchar *text = NULL;
    gint npoint = 0;
 
    while(*aux){
       child = *aux++;
       value = *aux++;

       if(strcmp(child, "Text") == 0) text = value; 
       if(strcmp(child, "Point") == 0) npoint = atof(value); 
    }

    if(npoint < state->dataset->real_data->num_points)
       state->dataset->real_data->labels[npoint] = g_strdup(text);
  }

/***********************************************************
 * LINES
 ***********************************************************/
  if(strcmp(name, "Line") == 0){
    gint style = 0;
    gint width = 0;
    GdkColor color;
 
    while(*aux){
       child = *aux++;
       value = *aux++;

       if(strcmp(child, "Style") == 0) style = atoi(value); 
       if(strcmp(child, "Width") == 0) width = atoi(value); 
       if(strcmp(child, "R") == 0) color.red = atoi(value); 
       if(strcmp(child, "G") == 0) color.green = atoi(value); 
       if(strcmp(child, "B") == 0) color.blue = atoi(value); 
    }
    gdk_color_alloc(gdk_colormap_get_system(), &color);

    switch(state->state){
        case PARSER_IN_AXIS:
          gtk_plot_axis_set_attributes(real_plot, state->axis,
                                       width, &color);
	  break;
        case PARSER_IN_VMAJOR:
          gtk_plot_major_vgrid_set_attributes(real_plot, style, width, &color);
	  break;
        case PARSER_IN_VMINOR:
          gtk_plot_minor_vgrid_set_attributes(real_plot, style, width, &color);
	  break;
        case PARSER_IN_HMAJOR:
          gtk_plot_major_hgrid_set_attributes(real_plot, style, width, &color);
	  break;
        case PARSER_IN_HMINOR:
          gtk_plot_minor_hgrid_set_attributes(real_plot, style, width, &color);
	  break;
        case PARSER_IN_X0:
          gtk_plot_x0line_set_attributes(real_plot, style, width, &color);
	  break;
        case PARSER_IN_Y0:
          gtk_plot_y0line_set_attributes(real_plot, style, width, &color);
	  break;
        case PARSER_IN_DATA:
          gtk_plot_data_set_line_attributes(state->dataset->real_data,
                                            style, width, &color);
	  break;
        case PARSER_IN_XLINE:
          gtk_plot_data_set_x_attributes(state->dataset->real_data,
                                         style, width, &color);
	  break;
        case PARSER_IN_YLINE:
          gtk_plot_data_set_y_attributes(state->dataset->real_data,
                                         style, width, &color);
	  break;
        default:
	  break;
    }

  }

/***********************************************************
 * SYMBOLS 
 ***********************************************************/
  if(strcmp(name, "Symbol") == 0){
    gint type = 0;
    gint style = 0;
    gint size = 0;
    gint width = 0;
    GdkColor color;
 
    while(*aux){
       child = *aux++;
       value = *aux++;

       if(strcmp(child, "Type") == 0) type = atoi(value); 
       if(strcmp(child, "Style") == 0) style = atoi(value); 
       if(strcmp(child, "Size") == 0) size = atoi(value); 
       if(strcmp(child, "Width") == 0) width = atoi(value); 
       if(strcmp(child, "R") == 0) color.red = atoi(value); 
       if(strcmp(child, "G") == 0) color.green = atoi(value); 
       if(strcmp(child, "B") == 0) color.blue = atoi(value); 
    }
    gdk_color_alloc(gdk_colormap_get_system(), &color);

    gtk_plot_data_set_symbol(state->dataset->real_data,
                             type, style, size, width, &color);
  }
/***********************************************************
 * TEXT 
 ***********************************************************/
  if(strcmp(name, "Text") == 0){
    gdouble x = 0.;
    gdouble y = 0.;
    gint angle = 0;
    gint justification = 0;
    gboolean transparent = TRUE;

    state->in_text = TRUE;
 
    while(*aux){
       child = *aux++;
       value = *aux++;

       if(strcmp(child, "X") == 0) x = atof(value); 
       if(strcmp(child, "Y") == 0) y = atof(value); 
       if(strcmp(child, "Angle") == 0) angle = atoi(value); 
       if(strcmp(child, "Justification") == 0) justification = atoi(value); 
       if(strcmp(child, "Transparent") == 0) transparent = atoi(value); 
    }

    state->text.x = x;
    state->text.y = y;
    state->text.angle = angle;
    state->text.justification = justification;
    state->text.transparent = transparent;
  }

  if(strcmp(name, "Foreground") == 0){
    GdkColor color;
 
    while(*aux){
       child = *aux++;
       value = *aux++;

       if(strcmp(child, "R") == 0) color.red = atoi(value); 
       if(strcmp(child, "G") == 0) color.green = atoi(value); 
       if(strcmp(child, "B") == 0) color.blue = atoi(value); 
    }
    gdk_color_alloc(gdk_colormap_get_system(), &color);

    if(state->in_text) state->text.fg = color;
  }

  if(strcmp(name, "Background") == 0){
    GdkColor color;
 
    while(*aux){
       child = *aux++;
       value = *aux++;

       if(strcmp(child, "R") == 0) color.red = atoi(value); 
       if(strcmp(child, "G") == 0) color.green = atoi(value); 
       if(strcmp(child, "B") == 0) color.blue = atoi(value); 
    }
    gdk_color_alloc(gdk_colormap_get_system(), &color);

    if(state->in_text) state->text.bg = color;
  }
 
  if(strcmp(name, "Font") == 0){
    const gchar *name = NULL; 
    gint height = 12;

    while(*aux){
       child = *aux++;
       value = *aux++;

       if(strcmp(child, "Name") == 0) name = value; 
       if(strcmp(child, "Height") == 0) height = atoi(value); 
    }

    if(state->text.font)
      g_free(state->text.font);

    state->text.font = g_strdup(name);
    state->text.height = height;
  }
 

}

static void
sgEndElement(sgParseState *state, const CHAR *real_name)
{
  SGworksheet *worksheet = state->worksheet;
  GtkSheet *sheet = NULL;
  const CHAR *name;

  name = get_real_name(real_name);

  if(worksheet) sheet = GTK_SHEET(worksheet->sheet);
  
  if(strcmp(name, "Worksheet") == 0){
     state->state = PARSER_UNKNOWN;
     state->layer = NULL;
  }

  if(strcmp(name, "MaxRow") == 0)
     sg_worksheet_add_rows(state->worksheet,
                           atoi(state->content->str)-sheet->maxrow+1);

  if(state->state == PARSER_IN_WORKSHEET && strcmp(name, "Name") == 0){
     sg_worksheet_rename(state->worksheet, state->content->str);
  }

  if(strcmp(name, "Begin") == 0)
     sg_worksheet_set_begin(state->worksheet, atoi(state->content->str));

  if(strcmp(name, "End") == 0)
     sg_worksheet_set_end(state->worksheet, atoi(state->content->str));

  if(strcmp(name, "Content") == 0){
     sg_worksheet_cell_set(state->worksheet, state->row, state->col,
                           state->content->str, FALSE);
  }

  if(strcmp(name, "Formula") == 0)
     sg_worksheet_column_set_exp(state->worksheet, state->col,
                                 state->content->str);

  if(state->state == PARSER_IN_PLOT && strcmp(name, "Name") == 0)
     sg_plot_rename(state->plot, state->content->str); 

  if(strcmp(name, "Plot") == 0){
     state->state = PARSER_UNKNOWN;
     state->num_layers = 0;
  }

  if(strcmp(name, "Layer") == 0){
     state->state = PARSER_IN_PLOT;
  }

  if(strcmp(name, "Vmajor") == 0){
     state->state = PARSER_IN_LAYER;
  }

  if(strcmp(name, "Vminor") == 0){
     state->state = PARSER_IN_LAYER;
  }

  if(strcmp(name, "Hmajor") == 0){
     state->state = PARSER_IN_LAYER;
  }

  if(strcmp(name, "Hminor") == 0){
     state->state = PARSER_IN_LAYER;
  }

  if(strcmp(name, "X0") == 0){
     state->state = PARSER_IN_LAYER;
  }

  if(strcmp(name, "Y0") == 0){
     state->state = PARSER_IN_LAYER;
  }

  if(strcmp(name, "Axis") == 0){
     state->state = PARSER_IN_LAYER;
  }

  if(strcmp(name, "Ticks") == 0){
     state->state = PARSER_IN_AXIS;
  }

  if(strcmp(name, "Title") == 0){
     state->state = PARSER_IN_AXIS;
  }

  if(strcmp(name, "Labels") == 0){
     state->state = PARSER_IN_AXIS;
  }

  if(strcmp(name, "Legends") == 0){
     state->state = PARSER_IN_LAYER;
  }

  if(strcmp(name, "Legend") == 0 && state->state == PARSER_IN_DATA){
     gtk_plot_data_set_legend(state->dataset->real_data, state->content->str); 
  }

  if(strcmp(name, "String") == 0 && state->in_text){
     state->text.text = state->content->str;
  }

  if(strcmp(name, "Text") == 0){
     switch(state->state){
       case PARSER_IN_PLOT:
         gtk_plot_canvas_put_text(GTK_PLOT_CANVAS(state->plot->real_canvas),
				  state->text.x,
				  state->text.y,
				  state->text.angle,
				  state->text.font,
				  state->text.height,
				  &state->text.fg,
				  &state->text.bg,
				  state->text.justification,
				  state->text.text);
         break;
       case PARSER_IN_LEGENDS:
         gtk_plot_legends_set_attributes(GTK_PLOT(state->layer->real_plot),
				         state->text.font,
				         state->text.height,
				         &state->text.fg,
				         &state->text.bg);
         break;
       case PARSER_IN_AXIS:
         gtk_plot_axis_labels_set_attributes(GTK_PLOT(state->layer->real_plot),
                                             state->axis,
			 	             state->text.font,
				             state->text.height,
				             state->text.angle,
				             &state->text.fg,
				             &state->text.bg);

         break;
       case PARSER_IN_TITLE:
         gtk_plot_axis_title_set_attributes(GTK_PLOT(state->layer->real_plot),
                                            state->axis,
				            state->text.font,
				            state->text.height,
				            state->text.angle,
				            &state->text.fg,
				            &state->text.bg);
         if(state->text.text && strlen(state->text.text) > 0){
           gtk_plot_axis_set_title(GTK_PLOT(state->layer->real_plot),
                                   state->axis,
                                   state->text.text);
         }
         break;
       case PARSER_IN_DATA:
         gtk_plot_data_labels_set_attributes(state->dataset->real_data,
				             state->text.font,
				             state->text.height,
				             state->text.angle,
				             &state->text.fg,
				             &state->text.bg);
         break;
       default:
         break;
     }

     state->in_text = FALSE;

     if(state->text.font)
        g_free(state->text.font);
     state->text.font = NULL;
     return;
  }

  if(strcmp(name, "Dataset") == 0){
     state->state = PARSER_IN_LAYER;
     sg_layer_add_dataset(state->layer, state->dataset);
  }

  if(strcmp(name, "Points") == 0){
     state->state = PARSER_IN_DATA;
  }

  if(strcmp(name, "Datalabels") == 0){
     state->state = PARSER_IN_DATA;
  }

  if(strcmp(name, "Xline") == 0){
     state->state = PARSER_IN_DATA;
  }

  if(strcmp(name, "Yline") == 0){
     state->state = PARSER_IN_DATA;
  }

  g_free(state->name);
  state->name = NULL;
}


void  
sg_project_file_import_xml(gchar *file, SGworksheet *worksheet, SGplot *plot)
{
  sgParseState state;

  state.worksheet = worksheet;
  state.plot = plot;
  state.layer = NULL;
  state.dataset = NULL;

  state.text.font = NULL;
  state.text.text = NULL;
  state.text.height = 12;

  state.in_text = FALSE;

  state.num_layers = 0;

  state.object = PARSE_PROJECT;
  if(worksheet) state.object = PARSE_WORKSHEET;
  if(plot) state.object = PARSE_PLOT;

  if(xmlSAXUserParseFile(&sgSAXParser, &state, file) < 0) {
       sg_message_dialog("Document not well formed!", 1);
       return;
  }
}

static const CHAR *
get_real_name(const CHAR *name)
{
  const CHAR *real_name = name;

  while(real_name && real_name != '\0'){
    if(*real_name++ == ':') return real_name;
  }

  return name;
} 
 
